// Missing: VMS_CMD_INT; Rest of ints.
	
// scmGfx.cpp - Implementation of the graphics part for the Scheme

#include "stdafx.h"
#include "Board.h"
#include "BoardView.h"
#include "MsgQ.h"
#include "comb.h"
#include "turtlemsg.h"
#include "turtle.h"
#include "threads.h"
#include "messages.h"

#include "Turtle.h"

#include "mydebug.h"
#include "scm.h"
extern "C" {
#include "setjump.h"
           }
#ifdef TC16_TURTLE		    
# define tc16_turtle TC16_TURTLE
#else
long tc16_turtle;
#endif

/* a window looks like ( tc16_window . ( name . CBoardView* ) ) */
/* a turtle looks like ( tc16_window . ( name .   CTurtle * ) ) */
#ifdef TC16_WINDOW
# define tc16_window TC16_WINDOW
#else
long tc16_window;
#endif 



#define TURTLEP(x)  (CELLP(x) && (TYP16(x) == tc16_turtle))
#define LVTRTLP(x) (CELLP(x) && (TYP16(x) == tc16_turtle) && (TRTLPTR(x)!=NULL))  /* live turtle */
#define WINDOWP(x)  (CELLP(x) && (TYP16(x) == tc16_window))
#define LVWINP(x)  (CELLP(x) && (TYP16(x) == tc16_window) && (WINPTR(x)!=NULL))  /* live window */
#define WINPTR(xy)    (CDR(CDR(xy)))
#define WINNAME(xy)   (CAR(CDR(xy)))
#define TRTLPTR(xy)   (CDR(CDR(xy)))
#define TRTLNAME(xy)  (CAR(CDR(xy)))


//inline msgqP(SCM x) { return NIMP(x) && MSGQP(x); }


static int __cdecl print_turtle(SCM exp, SCM port, int writing)
{
	lputs ("#<", port);
	if (TRTLPTR(exp)==0) lputs ("dead ", port);
	lputs ("turtle", port);
	if (!UNBNDP(TRTLNAME(exp)))
	{
		lputc(' ', port);
		iprin1(TRTLNAME(exp),port,writing);
	}
	else if (TRTLPTR(exp)!=0)
	{
		lputs(" 0x", port);
		intprint(TRTLPTR(exp), 16, port);
	}
	lputc('>', port);
	return !0;
}

// A turtle looks like that:

//  cel----->heapcell[tc16_turtle |  +  ]
//                                   |
//                             heapcell[  +  | CTurtle*  ]
//								         /      
//                      SCM turtle name	      
//
// So we have two heapocells to mark directly, plus the name which is marked by gc_mark()

extern "C" void mark_locations(STACKITEM x[], sizet n);
SCM __cdecl mark_turtle(SCM cel)
{
	if LVTRTLP(cel)
	{ // Mark interrupts of live turtles only
		if (((CTurtle *)(TRTLPTR(cel)))->m_pUserIntQ)  // there is a user interrupt
			mark_locations((STACKITEM *)(&(((CTurtle *)(TRTLPTR(cel)))->m_UserIntMsg)), sizeof DWORD / sizeof STACKITEM);
		
		Msgs Ms;
		
		if (((CTurtle *)(TRTLPTR(cel)))->ResetIterateInts())
		{
			while (((CTurtle *)(TRTLPTR(cel)))->IterateInts(&Ms))
			{ 
				if (Ms.Kind == BOTH_INTS || Ms.Kind == IS_ON)
					mark_locations((STACKITEM *)(&(Ms.MsgIs)), sizeof DWORD / sizeof STACKITEM);
				
				if (Ms.Kind == BOTH_INTS || Ms.Kind == NOT_ON)
					mark_locations((STACKITEM *)(&(Ms.MsgNot)), sizeof DWORD / sizeof STACKITEM);
			}
		}
	}
	
	SETGC8MARK(cel);			// smob heapcells are marked like this
	SETGCMARK(CDR(cel));		// cons heapcells are markes like this
	return CAR(CDR(cel));		// And go on and mark the name.  
}

static SCM __cdecl kill_turtle(SCM);

static sizet __cdecl free_turtle(SCM cel)
{
	if LVTRTLP(cel)
		( (CTurtle*)TRTLPTR(cel) ) -> nullDI();
	return 0;
}
	
static smobfuns turtle_smob = 
{ (SCM(__cdecl*)(SCM)) mark_turtle, 
  (sizet(__cdecl*)(CELLPTR)) free_turtle, 
  (int(__cdecl*)(SCM,SCM,int))print_turtle, 
  0 };

static int __cdecl print_window(SCM exp, SCM port, int writing)
{
	lputs("#<", port);
	if (!LVWINP(exp)) lputs ("dead ", port);
	lputs("window ", port);
	if (UNBNDP(WINNAME(exp)))
	{
		lputs("0x", port);
		intprint(WINPTR(exp), 16, port);
	}
	else
		iprin1(WINNAME(exp), port, writing);
	lputc('>', port);
	return !0;
}
 
//static SCM __cdecl mark_window(SCM cel) - same as mark_turtle
SCM __cdecl mark_window(SCM cel)
{
	if LVWINP(cel)
		mark_locations(
			(STACKITEM*)&(((CBoardView*)WINPTR(cel))->m_UserIntMsg),
			sizeof DWORD / sizeof STACKITEM);

	SETGC8MARK(cel);			// smob heapcells are marked like this
	SETGCMARK(CDR(cel));		// cons heapcells are markes like this
	return CAR(CDR(cel));		// And go on and mark the name.  
}

	
static sizet __cdecl free_window(SCM cel)
{
	if LVWINP(cel)
		((CBoardView*)WINPTR(cel))->nullDI();
	return 0;
}

static smobfuns window_smob = 
{ (SCM(__cdecl*)(SCM)) mark_window, 
  (sizet(__cdecl*)(CELLPTR)) free_window, 
  (int(__cdecl*)(SCM,SCM,int))print_window, 
  0 };

/* Posts a rename meesage to the GUI */

void PostUpdateName(SCM x)
{
	WASSERT (WINDOWP(x));
	
	CString *s;
	if (UNBNDP(WINNAME(x)))
	{
		s=new CString;
		s->Format("Graphic window 0x%X", WINPTR(x));
	}
	else
	{
		s=new CString(CHARS(WINNAME(x)), LENGTH(WINNAME(x)));
	}
	TheOutputQueue.Post(CMessage(MSG_SETNAME_GFXWIN, (DWORD)s, WINPTR(x)));
}



/* (new-window [name]) where name is a symbol or a string */
/* Creates a window, with its title set to name */
#define myname "new-window"
static SCM __cdecl create_window(/*optional*/SCM name)
{
	ASSERT(UNBNDP(name) || SYMBOLP(name) || STRINGP(name), name, ARG1, myname);

	CMessageQueue* pQ = new CMessageQueue;
	TheOutputQueue.Post(CMessage(MSG_TYPE_CREATE_VIEW, 0, pQ));
	CMessage msg;
	pQ->Get(msg);
	delete pQ;
	SCM x,y;
	DEFER_INTS;
	NEWCELL(x);
	NEWCELL(y);
	CAR(x)=tc16_window;
	CDR(x)=y;
	CAR(y)=name; /* WINNAME(x) */
	CDR(y)=(SCM)(msg.ptr); /* WINPTR(x) */
	ALLOW_INTS;
	
	PostUpdateName(x);
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_VIEW,  
		(DWORD)new CView_Msg(MSG_CMD_DEATH_PP, &CDR(y)),
		CDR(y)));
	
	return x;
}
#undef myname

#define myname "rename-window"
static SCM __cdecl rename_window(SCM window, /*optional*/SCM name)
{
	ASSERT(WINDOWP(window), window, ARG1, myname);
	ASSERT(UNBNDP(name) || SYMBOLP(name) || STRINGP(name), name, ARG2, myname);
	WINNAME(window)=name;
	PostUpdateName(window);
	return UNDEFINED;
}
#undef myname

/*case MSG_TYPE_NEW_TURTLE:
			((CBoardView *)(msg.ptr))->CreateTurtle((CMessageQueue *)(msg.msg));
		*/

#define myname "make-turtle"
SCM __cdecl make_turtle(SCM window, /*optional*/ SCM name)
{
	ASSERT(WINDOWP(window), window, ARG1, myname);
	CMessageQueue* pQ = new CMessageQueue;
	TheOutputQueue.Post(CMessage(MSG_TYPE_NEW_TURTLE, (DWORD)pQ, WINPTR(window)));
	CMessage msg;
	pQ->Get(msg);
	delete pQ;
	SCM x,y;
	DEFER_INTS;
	NEWCELL(x);
	NEWCELL(y);
	CAR(x)=tc16_turtle;
	CDR(x)=y;
	CAR(y)=name;            // TRTLNAME(x)
	CDR(y)=(SCM)(msg.ptr);  // TRTLPTR(x)
	ALLOW_INTS;
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_TURTLE, 
		(DWORD)new CTurtle_Msg(MSG_CMD_DEATH_PP, (DWORD)&CDR(y)),
		CDR(y)));
	return x;
}
#undef myname

#define myname "turtle-alive?"
SCM __cdecl turtle_alive_p(SCM x)
{
	ASSERT(TURTLEP(x), x, ARG1, myname);
	return (TRTLPTR(x)==NULL)? BOOL_F : BOOL_T ;
}
#undef myname

#define myname "window-alive?"
SCM __cdecl window_alive_p(SCM x)
{
	ASSERT(WINDOWP(x), x, ARG1, myname);
	return (LVWINP(x))? BOOL_T : BOOL_F ;
}
#undef myname

#define myname "clone-turtle"
SCM __cdecl clone_turtle(SCM turtle, /*optional*/ SCM name)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	CMessageQueue* pQ = new CMessageQueue;
	TheOutputQueue.Post(CMessage(MSG_TYPE_CLONE_TURTLE, (DWORD)pQ, TRTLPTR(turtle)));
	CMessage msg;
	pQ->Get(msg);
	SCM x,y;
	DEFER_INTS;
	NEWCELL(x);
	NEWCELL(y);
	CAR(x)=tc16_turtle;
	CDR(x)=y;
	CAR(y)=UNBNDP(name)? TRTLNAME(turtle) : name;
	CDR(y)=(SCM)(msg.ptr);   //TRTLPTR(x)
	ALLOW_INTS;
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_TURTLE, 
		(DWORD)new CTurtle_Msg(MSG_CMD_DEATH_PP, (DWORD)&CDR(y)),
		CDR(y)));
	return x;
}
#undef myname

// rename a turtle. If the optional name is ommited, then the name is removed.
#define myname "rename-turtle"
static SCM __cdecl rename_turtle(SCM ttl, /*optional*/SCM name)
{
	ASSERT(TURTLEP(ttl), ttl, ARG1, myname); //Even dead turtles allowed.
	TRTLNAME(ttl)=name;
	return UNDEFINED;
}
#undef myname

#define myname "turtle-name"
static SCM __cdecl turtle_name(SCM ttl)
{
	ASSERT(TURTLEP(ttl), ttl, ARG1, myname); //Even dead turtles allowed.
	return TRTLNAME(ttl);
}
#undef myname

#define myname "window-name"
static SCM __cdecl window_name(SCM w)
{
	ASSERT(WINDOWP(w), w, ARG1, myname); 
	return WINNAME(w);
}
#undef myname

#define myname "forward!"
SCM __cdecl turtle_forward(SCM turtle, SCM d)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	unsigned long dist=num2long(d, (char*)ARG2, myname);
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE, (DWORD)new CTurtle_Msg(MSG_CMD_FD, dist), TRTLPTR(turtle) ));
	return UNDEFINED;
}
#undef myname

#define myname "turtle-width!"
SCM __cdecl turtle_width(SCM turtle, SCM d)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	unsigned long w=num2long(d, (char*)ARG2, myname);
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE, (DWORD)new CTurtle_Msg(MSG_CMD_WIDTH, w), TRTLPTR(turtle) ));
	return UNDEFINED;
}
#undef myname

#define myname "backwards!"
SCM __cdecl turtle_backwards(SCM turtle, SCM d)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	unsigned long dist=-num2long(d, (char*)ARG2, myname);
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE, (DWORD)new CTurtle_Msg(MSG_CMD_FD, dist), TRTLPTR(turtle) ));
	return UNDEFINED;
}
#undef myname

#define myname "right!"
SCM __cdecl turtle_right(SCM turtle, SCM a)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	unsigned long angle=num2long(a, (char*)ARG2, myname);
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE, (DWORD)new CTurtle_Msg(MSG_CMD_RIGHT, angle), TRTLPTR(turtle) ));
	return UNDEFINED;
}
#undef myname

#define myname "left!"
SCM __cdecl turtle_left(SCM turtle, SCM a)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	unsigned long angle=num2long(a, (char*)ARG2, myname);
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE, (DWORD)new CTurtle_Msg(MSG_CMD_LEFT, angle), TRTLPTR(turtle) ));
	return UNDEFINED;
}
#undef myname

#define myname "set-heading!"
SCM __cdecl set_heading(SCM turtle, SCM d)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	long ang=num2long(d, (char*)ARG2, myname);
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE, 
		(DWORD)new CTurtle_Msg(MSG_CMD_HEADING, ang), 
		TRTLPTR(turtle) ));
	return UNDEFINED;
}
#undef myname

#define myname "move-to!"
SCM __cdecl move_to(SCM turtle, SCM newx, SCM newy)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	unsigned long x=num2long(newx, (char*)ARG2, myname);
	unsigned long y=num2long(newy, (char*)ARG3, myname);
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE,
		(DWORD)new CTurtle_Msg(MSG_CMD_MOVE_ABS,(DWORD)(new CPoint(x,y))),
		TRTLPTR(turtle)));
	return UNDEFINED;
}
#undef myname

#define myname "draw-to!"
SCM __cdecl draw_to(SCM turtle, SCM newx, SCM newy)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	unsigned long x=num2long(newx, (char*)ARG2, myname);
	unsigned long y=num2long(newy, (char*)ARG3, myname);
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE,
		(DWORD)new CTurtle_Msg(MSG_CMD_GLIDE,(DWORD)(new CPoint(x,y))),
		TRTLPTR(turtle)));
	return UNDEFINED;
}
#undef myname





#define myname "move-turtle-to-window"   /* (move-turtle-to-window t w) */
SCM __cdecl move_turtle_to_window(SCM turtle, SCM window)
{
	ASSERT(LVTRTLP(turtle), turtle, ARG1, myname);
	ASSERT(WINDOWP(window), window, ARG2, myname);
	
	TheOutputQueue.Post(
		CMessage(
		MSG_TYPE_TURTLE, 
		(DWORD)new CTurtle_Msg(MSG_CMD_MOVE, WINPTR(window)), 
		TRTLPTR(turtle) ));
	return UNDEFINED;
}
#undef myname

#define META1(scm_fn, c_fn, msgtyp)            \
static SCM __cdecl c_fn(SCM ttl)               \
{	ASSERT(LVTRTLP(ttl), ttl, ARG1, scm_fn);   \
	TheOutputQueue.Post(                       \
		CMessage(MSG_TYPE_TURTLE,              \
		(DWORD)new CTurtle_Msg(msgtyp, 0),     \
		TRTLPTR(ttl)));                        \
	return UNDEFINED;                          \
}

META1("pen-up!",				pen_up,				MSG_CMD_PENUP)
META1("pen-down!",				pen_down,			MSG_CMD_PENDOWN)
META1("pen-erase!",				pen_erase,			MSG_CMD_ERASE)
META1("show-turtle!",			show_turtle,		MSG_CMD_SHOW)
META1("hide-turtle!",			hide_turtle,		MSG_CMD_HIDE)
META1("make-turtle-solid!",		make_turtle_solid,	MSG_CMD_SOLID)
META1("make-turtle-hollow!",	make_turtle_hollow,	MSG_CMD_HOLLOW)

#define myname "kill-turtle"
static SCM __cdecl kill_turtle(SCM ttl) 
{	ASSERT(LVTRTLP(ttl), ttl, ARG1, myname);  
	TheOutputQueue.Post(CMessage(MSG_TYPE_KILL_TURTLE, 0, TRTLPTR(ttl)));
	TRTLPTR(ttl)=NULL;
	return UNDEFINED;  
}
#undef myname

#define myname "set-color!"
// (set-color! turtle r g b) or (set-color! turtle (list r g b))
SCM __cdecl set_color(SCM args)
{
	SCM x=args;
	int r,g,b=ilength(args);

	ASSERT(b==4 || b==2, args, WNA, myname);
	ASSERT(LVTRTLP(CAR(x)), CAR(x), ARG1, myname);
	if (b==4)
	{	
		r=num2long(CAR(x=CDR(x)), (char*)ARG2, myname);
		g=num2long(CAR(x=CDR(x)), (char*)ARG3, myname);
		b=num2long(CAR(  CDR(x)), (char*)ARG4, myname);
	}
	else
	{
		x=CAR(CDR(args));
		ASSERT(CONSP(x), x, (char*)ARG2, myname);
		b=ilength(x);
		ASSERT(b==3, x, (char*)ARG2, myname);
		r=num2long(CAR(x       ), (char*)ARG2, myname);
		g=num2long(CAR(x=CDR(x)), (char*)ARG2, myname);
		b=num2long(CAR(  CDR(x)), (char*)ARG2, myname);
	}
		
	TheOutputQueue.Post(  
		CMessage(
		MSG_TYPE_TURTLE, 
		(DWORD)new CTurtle_Msg(MSG_CMD_COLOR, RGB(r,g,b)), 
		TRTLPTR(CAR(args))));
	return UNDEFINED;
}
#undef myname

#define META_INNERSTATE(scm_fn_name,c_fn_name,retval_expr)            \
SCM __cdecl c_fn_name(SCM turtle)									  \
{																	  \
	ASSERT(LVTRTLP(turtle), turtle, ARG1, scm_fn_name);				  \
	CMessageQueue* pQ=new CMessageQueue;							  \
	TheOutputQueue.Post(CMessage(									  \
		MSG_TYPE_TURTLE,											  \
		(DWORD)new CTurtle_Msg(MSG_ASK_INNER, (DWORD)pQ),			  \
		TRTLPTR(turtle)));											  \
	CMessage msg;													  \
	pQ->Get(msg);													  \
	delete pQ;														  \
	CTurtleInnerState *pTIS= (CTurtleInnerState*)(msg.msg);			  \
	DEFER_INTS;														  \
	SCM retval=retval_expr;											  \
	ALLOW_INTS;														  \
	delete pTIS;													  \
	return retval;           										  \
}


#define BOOL2SCM(k) ( (k)? BOOL_T:BOOL_F )

#define LIST6(a,b,c,d,e,f) ( cons2(a,b,cons2(c,d,cons2(e,f, EOL))) )
#define LIST3(a,b,c) ( cons2(a,b,cons(c, EOL)) )

SCM colorref2list(COLORREF cr)
{ return LIST3(
			   MAKINUM(GetRValue(cr)), 
			   MAKINUM(GetGValue(cr)), 
			   MAKINUM(GetBValue(cr)));
}

SCM dbl2scm(double d, char* caller)
{
	SCM f;
	NEWCELL(f);
	CDR(f) = (SCM)must_malloc(1L*sizeof(double), caller);
	REAL(f) = d;
	CAR(f) = tc_dblr;
	return f;
}

SCM cpoint2pair(CPoint pt, char* caller)
	{ return cons(dbl2scm(pt.x, caller), dbl2scm(pt.y, caller)); }

SCM ptis2list(CTurtleInnerState* pTIS, char* caller)
{
	return LIST6(
		colorref2list(pTIS->Color),
		dbl2scm(pTIS->Heading, caller),
		BOOL2SCM(!pTIS->Hidden),
		MAKINUM(pTIS->PenWidth),
		BOOL2SCM(pTIS->IsSolid),
		cpoint2pair(pTIS->Location, caller) );
}

META_INNERSTATE("turtle-color", ist_color, colorref2list(pTIS->Color))
META_INNERSTATE("turtle-heading", ist_heading, dbl2scm(pTIS->Heading, "turtle-heading"))
META_INNERSTATE("turtle-visible", ist_visible, BOOL2SCM(!pTIS->Hidden))
META_INNERSTATE("turtle-width", ist_width, MAKINUM(pTIS->PenWidth))
META_INNERSTATE("turtle-solid?", ist_solid, BOOL2SCM(pTIS->IsSolid))
META_INNERSTATE("turtle-position", ist_pos, cpoint2pair(pTIS->Location, "turtle-position"))
META_INNERSTATE("turtle-inner-state", ist_state, ptis2list(pTIS, "turtle-inner-state"))
// elad! missing "pen-is-up, down, erase"

COLORREF list2rgb(SCM x, char* pos, char*caller)
{
	ASSERT(ilength(x)==3, x, pos, caller);
	return	RGB(
		num2long(CAR(x),             pos, caller),
		num2long(CAR(CDR(x)),        pos, caller),
		num2long(CAR(CDR(CDR(x))),   pos, caller) );
}


// ------- SEEING primtives -----------

// (look t distance angle '(r g b))   - Look for color
// (look t distance angle [t1]) - Look for (any) turtle

#define myname "look"
SCM __cdecl look4(SCM args) 
{
	int i = ilength(args);
	ASSERT(i==3 || i==4, args, WNA, myname);
	SCM trtl=CAR(args); args=CDR(args);
	ASSERT(LVTRTLP(trtl), trtl, ARG1, myname);
	UINT dist=num2ulong(CAR(args), (char*)ARG2, myname);
	args=CDR(args);
	double deg=num2dbl(CAR(args), (char*)ARG3, myname);
	args=CDR(args);
	CTurtleLookArgs tla;
	
	tla.Distance=dist;
	tla.Degree = deg;
	int msgtyp = MSG_CMD_LOOK4T;
    if NULLP(args)
		tla.Arg = (DWORD)NULL;
	else 
	{
		args=CAR(args);
		if LVTRTLP(args) tla.Arg=(DWORD)TRTLPTR(args);
		else
		{
			tla.Arg = list2rgb(args, (char*)ARG1, myname);
			msgtyp=MSG_CMD_LOOK4C;
		}
	}
	CMessageQueue* pQ;
	tla.AnsQueue = pQ = new CMessageQueue;
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_TURTLE,
		(DWORD) new CTurtle_Msg (msgtyp, (DWORD)new CTurtleLookArgs(tla)),
		TRTLPTR(trtl) ));
	CMessage msg;
	pQ->Get(msg);
	i=msg.msg;
	delete pQ;
	return BOOL2SCM(i);
}
#undef myname

// (turtle-notify t msg1 msg2 distange angle TARGET)
// Where TARGET is a color, a turtle, or missing for anyturtle
#define myname "turtle-notify"
SCM __cdecl turtle_watch (SCM args)
{
	int i=ilength(args);
	ASSERT(i==5 || i==6, args, WNA, myname);
	SCM ttl=CAR(args); 
	ASSERT(LVTRTLP(ttl), ttl, ARG1, myname);
	SCM IsMsg=CAR(args=CDR(args));
	SCM IsNotMsg=CAR(args=CDR(args));
	UINT Distance=num2ulong(CAR(args=CDR(args)), (char*)ARG4, myname);
	double Angle=num2dbl(CAR(args=CDR(args)), (char*)ARG5, myname);
	args=CDR(args);
	CTurtleInterrupt* pti;
	if NULLP(args) // Watch all turtles
		pti=new CTurtleInterrupt(
		Distance, Angle, (CTurtle*)NULL, IsMsg, 
		IsNotMsg, (CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue);
	else
	{
		args=CAR(args);
		if TURTLEP(args)  // Watch for a specific turtle
			pti=new CTurtleInterrupt(
			Distance, Angle, (CTurtle*)TRTLPTR(args), IsMsg, 
			IsNotMsg, (CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue);
		else 
		{ // Watch for a specific color
			COLORREF kolor = list2rgb(args, ARGn, myname);
			pti=new CTurtleInterrupt(
				Distance, Angle, kolor, IsMsg, 
				IsNotMsg, (CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue);
		}
	}
	TheOutputQueue.Post(CMessage
		(MSG_TYPE_TURTLE, 
		(DWORD)new CTurtle_Msg(MSG_CMD_INTERRUPT, (DWORD)pti), 
		TRTLPTR(ttl)));
	return UNSPECIFIED;
}
#undef myname

// (turtle-no-notify t distange angle TARGET)
// Where TARGET is a color, a turtle, or missing for anyturtle
#define myname "turtle-notify"
SCM __cdecl turtle_no_watch (SCM args)
{
	int i=ilength(args);
	ASSERT(i==5 || i==6, args, WNA, myname);
	SCM ttl=CAR(args); 
	ASSERT(LVTRTLP(ttl), ttl, ARG1, myname);
	UINT Distance=num2ulong(CAR(args=CDR(args)), (char*)ARG2, myname);
	double Angle=num2dbl(CAR(args=CDR(args)), (char*)ARG5, myname);
	args=CDR(args);
	CTurtleInterrupt* pti;
	if NULLP(args) // Watch all turtles
		pti=new CTurtleInterrupt(
		Distance, Angle, (CTurtle*)NULL, NULL, NULL,
		(CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue);
	else
	{
		args=CAR(args);
		if TURTLEP(args)  // Watch for a specific turtle
			pti=new CTurtleInterrupt(
			Distance, Angle, (CTurtle*)TRTLPTR(args), NULL, NULL,
			(CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue);
		else 
		{ // Watch for a specific color
			COLORREF kolor = list2rgb(args, ARGn, myname);
			pti=new CTurtleInterrupt(
				Distance, Angle, kolor, NULL,NULL,
				(CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue);
		}
	}
	TheOutputQueue.Post(CMessage
		(MSG_TYPE_TURTLE, 
		(DWORD)new CTurtle_Msg(MSG_CMD_DEL_INT, (DWORD)pti), 
		TRTLPTR(ttl)));
	return UNSPECIFIED;
}
#undef myname

void notification(SCM args, UINT which, char* myname)
{
	int i=ilength(args);
	ASSERT(i==4 || i==5, args, WNA, myname);
	SCM ttl=CAR(args); 
	ASSERT(LVTRTLP(ttl), ttl, ARG1, myname);
	SCM Msg=CAR(args=CDR(args));
	UINT Distance=num2ulong(CAR(args=CDR(args)), (char*)ARG3, myname);
	double Angle=num2dbl(CAR(args=CDR(args)), (char*)ARG4, myname);
	args=CDR(args);
	CTurtleInterrupt* pti;
	if NULLP(args) // Watch all turtles
		pti=new CTurtleInterrupt(
		Distance, Angle, (CTurtle*)NULL, Msg, NULL,
		(CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue, which);
	else
	{
		args=CAR(args);
		if TURTLEP(args)  // Watch for a specific turtle
			pti=new CTurtleInterrupt(
			Distance, Angle, (CTurtle*)TRTLPTR(args), Msg, NULL,
			(CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue, which);
		else 
		{ // Watch for a specific color
			COLORREF kolor = list2rgb(args, ARGn, myname);
			pti=new CTurtleInterrupt(
				Distance, Angle, kolor, Msg, NULL,
				(CTurtle*)TRTLPTR(ttl), TheCombination.pTheInputQueue, which);
		}
	}
	TheOutputQueue.Post(CMessage
		(MSG_TYPE_TURTLE, 
		(DWORD)new CTurtle_Msg(MSG_CMD_INTERRUPT, (DWORD)pti), 
		TRTLPTR(ttl)));
}

// (notify-when[-not] t1 msg dist ang TARGET)
static SCM __cdecl notify_when (SCM args)
{
	notification(args, IS_ON, "notify-when");
	return UNSPECIFIED;
}

static SCM __cdecl notify_when_not (SCM args)
{
	notification(args, NOT_ON, "notify-when-not");
	return UNSPECIFIED;
}

// ------------- Dealing with "interrupts" (UI UrgentMessages) -----

static SCM __cdecl clear_ints(SCM ttl)
{
	ASSERT(LVTRTLP(ttl), ttl, ARG1, "stop-notifying");
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_TURTLE,
		(DWORD) new CTurtle_Msg(MSG_CMD_CLR_INTS),
		(DWORD)TRTLPTR(ttl)));
	return UNDEFINED;
}

static SCM __cdecl resume_ints(SCM ttl)
{
	ASSERT(LVTRTLP(ttl), ttl, ARG1, "continue-notifying");
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_TURTLE,
		(DWORD)new CTurtle_Msg(MSG_CMD_SET_INTS),
		(DWORD)TRTLPTR(ttl)));
	return UNDEFINED;
}

static SCM __cdecl del_ints(SCM ttl)
{
	ASSERT(LVTRTLP(ttl), ttl, ARG1, "no-notifications");
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_TURTLE,
		(DWORD)new CTurtle_Msg(MSG_CMD_DEL_ALL_INTS),
		(DWORD)TRTLPTR(ttl)));
	return UNDEFINED;
}

// (notify-on-click turtle msg) 
// (notify-on-click window msg)
#define myname "notify-on-click"
static SCM __cdecl set_uium(SCM ttl, SCM um)
{
	if LVTRTLP(ttl)
		TheOutputQueue.Post(CMessage(
		MSG_TYPE_TURTLE,
		(DWORD)new CTurtle_Msg(MSG_CMD_USER_INT, (DWORD)um, (LPVOID)TheCombination.pTheInputQueue),
		(DWORD)TRTLPTR(ttl)));
	else if LVWINP(ttl)
		TheOutputQueue.Post(CMessage(
		MSG_TYPE_VIEW,
		(DWORD)new CView_Msg(MSG_CMD_USER_INT, (LPVOID)um, (LPVOID)TheCombination.pTheInputQueue),
		(DWORD)WINPTR(ttl)));
	else
		ASSERT(0, ttl, ARG1, myname);
	return UNSPECIFIED;
}
#undef myname

#define myname "clear-notify-click"
static SCM __cdecl clear_uium(SCM ttl)
{
	if LVTRTLP(ttl)
		TheOutputQueue.Post(CMessage(
		MSG_TYPE_TURTLE,
		(DWORD)new CTurtle_Msg(MSG_CMD_DEL_USER_INT),
		(DWORD)TRTLPTR(ttl)));
	else if LVWINP(ttl)
		TheOutputQueue.Post(CMessage(
		MSG_TYPE_VIEW,
		(DWORD)new CView_Msg(MSG_CMD_USER_INT),
		(DWORD)WINPTR(ttl)));
	else
		ASSERT(0, ttl, ARG1, myname);
	return UNSPECIFIED;
}
#undef myname






#define myname "set-background-color!" /* (of a view) */
SCM __cdecl set_bkgnd_color(SCM args)
{
	int i=ilength(args);
	ASSERT(i==2 || i==4, args, WNA, myname);
	ASSERT(WINDOWP(CAR(args)), CAR(args), ARG1, myname);
	COLORREF kolor;
	if (i==2) kolor=list2rgb(CAR(CDR(args)), (char*)ARG2, myname);
	else kolor=list2rgb(CDR(args), (char*)ARGn, myname);
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_VIEW,
		(DWORD)new CView_Msg(VMSG_CMD_SET_COLOR, (LPVOID)kolor),
		WINPTR(CAR(args))));
	return UNDEFINED;
}
#undef myname

#define myname "clear-window"
SCM __cdecl clear_window(SCM wind)
{
	ASSERT(WINDOWP(wind), wind, ARG1, myname);
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_VIEW,
		(DWORD)new CView_Msg(VMSG_CMD_CLS, /*uninteresting*/0),
		WINPTR(wind)));
	return UNDEFINED;
}
#undef myname

// (load-bitmap w "filename" [x y] )
#define myname "load-bitmap"
SCM __cdecl load_bitmap (SCM args)
{
	int i=ilength(args);
	ASSERT (i==2 || i==4, args, WNA, myname);
	ASSERT (WINDOWP(CAR(args)), CAR(args), ARG1, myname);
	ASSERT (STRINGP(CAR(CDR(args))), CAR(CDR(args)), ARG2, myname);
	CPoint DestLoc;
	BOOL IsMiddle=FALSE;
	if (i==2) IsMiddle=TRUE;
	else DestLoc=CPoint(num2ulong(CAR(CDR(CDR(args))), (char*)ARG3, myname),
		                num2ulong(CAR(CDR(CDR(CDR(args)))), (char*)ARG4, myname));
	CMessageQueue* pQ=new CMessageQueue;
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_VIEW,
		(DWORD)new CView_Msg(
			VMSG_CMD_OPEN_FILE, 
			new CViewOpenFileParams(
				CString(CHARS(CAR(CDR(args)))), 
				DestLoc, 
				IsMiddle,
				pQ)),
		WINPTR(CAR(args))));
	CMessage msg;
	pQ->Get(msg);
	BOOL bool=msg.msg;
	delete pQ;
	return bool==MSG_FILE_OPEN_OK?BOOL_T:BOOL_F;

}
#undef myname

// (save-bitmap w "filename")
#define myname "save-bitmap"
SCM __cdecl save_bitmap (SCM wind, SCM fname)
{
	ASSERT (WINDOWP(wind), wind,  ARG1, myname);
	ASSERT (STRINGP(fname), fname, ARG2, myname);
	CMessageQueue* pQ=new CMessageQueue;
	CString *pFileName = new CString(CHARS(fname));
	TheOutputQueue.Post(CMessage(
		MSG_TYPE_VIEW,
		(DWORD)new CView_Msg(
			VMSG_CMD_SAVE_FILE, 
		    LPVOID(pFileName),
			pQ),
			WINPTR(wind)));
	CMessage msg;
	pQ->Get(msg);
	BOOL bool=msg.msg;
	delete pQ;
	return bool==MSG_FILE_SAVE_OK?BOOL_T:BOOL_F;
}
#undef myname





SCM __cdecl turtlep(SCM x) { return TURTLEP(x)?BOOL_T:BOOL_F; }

SCM __cdecl windowp(SCM x) { return WINDOWP(x)?BOOL_T:BOOL_F; }


#define FIXF(_f) ( (SCM(__cdecl*)())(_f) )

static iproc subr1os[] = {
	{"new-window", FIXF(create_window)},
	{0,0} };

	
static iproc subr2os[] = {
	{"make-turtle",        FIXF(make_turtle)},
	{"clone-turtle",       FIXF(clone_turtle)},
	{"rename-turtle!",     FIXF(rename_turtle)},
	{"rename-window!",     FIXF(rename_window)},
	{"clear-window",       FIXF(clear_window)},
	{0,0}};

static iproc lsubrs[] = {
	{"set-color!", FIXF(set_color)},
	{"look", FIXF(look4)},
	{"turtle-notify", FIXF(turtle_watch)},
	{"turtle-no-notify", FIXF(turtle_no_watch)},
	{"notify-when", FIXF(notify_when)},	
	{"notify-when-not", FIXF(notify_when_not)},
	{"load-bitmap", FIXF(load_bitmap)},
	{"set-background-color!", FIXF(set_bkgnd_color)},
	{0,0}};

static iproc subr1s[] = {
	{"window?",				FIXF(windowp)},
	{"window-name",			FIXF(window_name)},
	{"window-alive?",		FIXF(window_alive_p)},
	{"turtle?",			    FIXF(turtlep)},
	{"turtle-alive?",		FIXF(turtle_alive_p)},
	{"turtle-name",			FIXF(turtle_name)},
	{"pen-up!",				FIXF(pen_up)},
	{"pen-erase!",			FIXF(pen_erase)},
	{"pen-down!",			FIXF(pen_down)},
	{"show-turtle!",		FIXF(show_turtle)},
	{"hide-turtle!",		FIXF(hide_turtle)},
	{"kill-turtle!",		FIXF(kill_turtle)},
	{"make-turtle-solid!",  FIXF(make_turtle_solid)},
	{"make-turtle-hollow!", FIXF(make_turtle_hollow)},
	{"turtle-solid?",      FIXF(ist_solid)},
	{"turtle-color",       FIXF(ist_color)},
	{"turtle-heading",     FIXF(ist_heading)},
	{"turtle-visible",     FIXF(ist_visible)},
	{"turtle-width",       FIXF(ist_width)},
	{"turtle-position",    FIXF(ist_pos)},
	{"turtle-inner-state", FIXF(ist_state)},
	{"stop-notifying",     FIXF(clear_ints)},
	{"continue-notifying", FIXF(resume_ints)},
	{"no-notification",    FIXF(del_ints)},
	{"clear-notify-click", FIXF(clear_uium)},
	{0,0} };

	static iproc subr2s[] = {
	{"forward!", FIXF(turtle_forward)},
	{"backwards!", FIXF(turtle_backwards)},
	{"left!", FIXF(turtle_left)},
	{"right!", FIXF(turtle_right)},
	{"set-heading!", FIXF(set_heading)},
	{"move-turtle-to-window", FIXF(move_turtle_to_window)},
	{"turtle-width!", FIXF(turtle_width)},
	{"notify-on-click", FIXF(set_uium)},
	{"save-bitmap", FIXF(save_bitmap)},
	{0,0} };

	static iproc subr3s[] = {
	{"move-to!", FIXF(move_to)},
	{"draw-to!", FIXF(draw_to)},
	{0,0} };

extern "C" void init_gfx()
{ 
#ifndef TC16_TURTLE
	tc16_turtle = 
#endif
		newsmob (&turtle_smob); 
#ifndef TC16_WINDOW
	tc16_window = 
#endif
		newsmob (&window_smob);
	init_iprocs(subr1s, tc7_subr_1);
	init_iprocs(subr2s, tc7_subr_2);
	init_iprocs(subr3s, tc7_subr_3);
	init_iprocs(subr1os, tc7_subr_1o);
	init_iprocs(subr2os, tc7_subr_2o);
	init_iprocs(lsubrs, tc7_lsubr);
}
